#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>

#include "../include/fio.h"
#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/disk.h"

#include "sfm.h"

#include "obj.h"
#include "simutils.h"
#include "menu.h"
#include "sarreality.h"
#include "weather.h"
#include "mission.h"
#include "sar.h"


char *SARMissionFSeekGetOpString(FILE *fp);

void SARMissionDestroy(sar_mission_struct *mission);

int SARMissionGetInterceptNameFromScene(
	const char *filename,
	const char *obj_name,
        sar_obj_intercept_struct *intercept
);

int SARMissionLoadMenpMapSceneInfo(
	sar_core_struct *core_ptr,
	sar_menu_struct *m,
	sar_menu_map_struct *map_ptr,
	const char *filename
);

char *SARMissionLoadDescription(const char *filename);
void SARMissionLoadMenuMap(
	sar_core_struct *core_ptr,
	sar_menu_struct *m,
	const char *filename
);

sar_mission_struct *SARMissionLoadFromFile(
	sar_core_struct *core_ptr,
	const char *filename,
	void *client_data,
	int (*progress_func)(void *, long, long)
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))

#define ISCOMMENT(c)    ((c) == '#')
#define ISCR(c)         (((c) == '\n') || ((c) == '\r'))


/*
 *	Return a statically allocated string indicating the
 *      operation fetched from the pointed to fp.  Return can be
 *      an empty string if there was no op string to be found or
 *      error.
 *
 *      fp is positioned at the end of the op string which the next
 *      fetch can be used to get its argument. If a new line character
 *      is detected at the end of string on file, then the fp will be
 *      repositioned at that new line character. The next reading of fp
 *      will read that new line character.
 */
char *SARMissionFSeekGetOpString(FILE *fp)
{
        int c, i;
#define len     80
        static char rtn_str[len];


        (*rtn_str) = '\0';

        if(fp == NULL)
            return(rtn_str);

        /* Seek past spaces. */
        FSeekPastSpaces(fp);

	/* Itterate to get parameter. */
        for(i = 0; i < len; i++)
        {
            c = fgetc(fp);
            if((c == EOF) ||
               ISBLANK(c)
            )
            {
                rtn_str[i] = '\0';
                break;
            }
            /* Escape sequence? */
            else if(c == '\\')
            {
                c = fgetc(fp);
                if(c == EOF)
                {
                    rtn_str[i] = '\0';
                    break;
                }

                if(c != '\\')
                    c = fgetc(fp);

		if(c == EOF)
                {
                    rtn_str[i] = '\0';
                    break;
                }
            }
	    /* New line? */
            else if(ISCR(c))
            {
                /* New line right at the end of the op string, seek   
                 * back one character.
                 */
                fseek(fp, -1, SEEK_CUR);  
   
                rtn_str[i] = '\0';
                break;
            }

            rtn_str[i] = (char)c;
        }
        
#undef len
        return(rtn_str);
}


/*
 *	Deallocates mission structure and all its substructures,
 *	including the mission structure itself.
 */
void SARMissionDestroy(sar_mission_struct *mission)
{
	sar_mobj_arrive_struct *arrive_objective;
	sar_mobj_rescue_struct *rescue_objective;

	if(mission == NULL)
	    return;

	switch(mission->type)
	{
	  case MISSION_TYPE_ARRIVE:
	    arrive_objective = mission->objective;
	    if(arrive_objective == NULL)
		break;

	    free(arrive_objective->arrive_at_name);
            free(arrive_objective->message_success);
            free(arrive_objective->message_fail);
	    break;

          case MISSION_TYPE_RESCUE:
            rescue_objective = mission->objective;
            if(rescue_objective == NULL)
                break;

            free(rescue_objective->arrive_at_name);
            free(rescue_objective->message_success);
            free(rescue_objective->message_fail);
            break;
	}


	free(mission->title);
	free(mission->description);

	free(mission->scene_file);
        free(mission->player_model_file);
        free(mission->player_stats_file);

        free(mission->objective);

	free(mission);

	return;
}

/*
 *	Load intercept by name from scene file, returns 0 on success
 *	or -1 on no match.
 */
int SARMissionGetInterceptNameFromScene(
	const char *filename,
	const char *obj_name,
	sar_obj_intercept_struct *intercept
)
{
	int c;
	FILE *fp;
	char *strptr;
	char *cur_obj_name = NULL;
	double value[10];

	int matched_translate = 0, matched_rotate = 0;


	if((filename == NULL) ||
           (obj_name == NULL) ||
	   (intercept == NULL)
	)
	    return(-1);

	fp = FOpen(filename, "rb");
	if(fp == NULL)
	    return(-1);

        /* Begin reading file. */
        while(1)
        {
            c = fgetc(fp);
            if(c == EOF)
                break;

            /* Read past leading spaces. */
            if(ISBLANK(c))
            {
                FSeekPastSpaces(fp);

                /* Get first non space character. */
                c = fgetc(fp);
                if(c == EOF)
                    break;
            }

            /* Newline character? */
            if(ISCR(c))
                continue;

            /* Comment? */
            if(ISCOMMENT(c))
            {
                FSeekNextLine(fp);	/* Seek to next line. */
                continue;
            }

            /* Seek back one character and fetch as operation. */
            fseek(fp, -1, SEEK_CUR);

            /* Get pointer to operation string. */
            strptr = SARMissionFSeekGetOpString(fp);

            /* Object name. */
            if(!strcasecmp(strptr, "object_name"))
            {
		free(cur_obj_name);
                cur_obj_name = FGetString(fp);

		matched_translate = 0;
		matched_rotate = 0;
            }
	    /* Translation. */
            else if(!strcasecmp(strptr, "translate") ||
                    !strcasecmp(strptr, "translation")
	    )
            {
		FGetValuesF(fp, value, 3);

		if((cur_obj_name != NULL) && !matched_translate)
		{
		    if(!strcasecmp(cur_obj_name, obj_name))
		    {
			intercept->x = value[0];
			intercept->y = value[1];
			intercept->z = SFMFeetToMeters(value[2]);

			intercept->radius = 40;	/* Explicitly set. */

			matched_translate = 1;
		    }
		}
            }
            /* Rotate. */
            else if(!strcasecmp(strptr, "rotate"))
            {
                FGetValuesF(fp, value, 3);

                if((cur_obj_name != NULL) && !matched_rotate)
                {
                    if(!strcasecmp(cur_obj_name, obj_name))
                    {


			matched_rotate = 1;
		    }
		}
            }
            /* Other. */
            else
            {
                FSeekNextLine(fp);
            }
	}

	/* Close file. */
	FClose(fp);

	/* Free names. */
	free(cur_obj_name);

	return(0);
}

/*
 *	Loads information from a scene file relivent to a mission map,
 *	such as map background file and landmarks.
 *
 *	Intercept points not loaded. Returns -1 on error or 0 on success.
 *
 *	This function is intended to be called by SARMissionLoadMenuMap().
 */
int SARMissionLoadMenpMapSceneInfo(
	sar_core_struct *core_ptr,
	sar_menu_struct *m,
	sar_menu_map_struct *map_ptr,
	const char *filename			/* Scene file. */
)
{
	int n;
        FILE *fp;
        char *strptr;
	sar_menu_map_marking_struct *marking_ptr;
	double value[10];


        if(filename == NULL)
            return(-1);

	/* Open scene file. */
        fp = FOpen(filename, "rb");
        if(fp == NULL)
            return(-1);

	/* Seek to parameter indicating the path to the image file
	 * representing the entire scene.
	 */
	rewind(fp);
	if(!FSeekToParm(fp, "entire_scene_map", SAR_COMMENT_CHAR, '\0'))
	{
            strptr = FGetString(fp);
            if(strptr != NULL)
            {
                char *strptr2;
                char **strv;
                int strc;
                char tmp_path[PATH_MAX + NAME_MAX];

                /* Explode arguments, format should be:
		 *
                 *	<width> <height> <file>
                 */
                strv = strexp(strptr, &strc);

		/* Don't need the original argument anymore. */
                free(strptr);
                strptr = NULL;


		/* Begin fetching values from exploded strings. */
		/* Width. */
                map_ptr->bg_tex_width = ((strc > 0) ?
                    atof(strv[0]) : 0
                );
		/* Height. */
                map_ptr->bg_tex_height = ((strc > 1) ?
                    atof(strv[1]) : 0
                );
		/* Image file path. */
                strptr = ((strc > 2) ? strv[2] : NULL);

                /* Complete the path to the image file. */
                if(strptr != NULL)   
                {
                    if(ISPATHABSOLUTE(strptr))
                    {           
                        strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                    }
                    else
                    {
                        struct stat stat_buf;

                        strptr2 = PrefixPaths(dname.local_data, strptr);
                        if(strptr2 != NULL)
                        {
                            if(stat(strptr2, &stat_buf))
                                strptr2 = PrefixPaths(
                                    dname.global_data,
                                    strptr
                                );
                        }
                        strncpy(
                            tmp_path,
                            (strptr2 == NULL) ? strptr : strptr2,
                            PATH_MAX + NAME_MAX
                        );
                    }

                    /* Unload menu map's background texture as needed. */
                    V3DTextureDestroy(map_ptr->bg_tex);

		    /* Load new texture for menu map's background. */
                    map_ptr->bg_tex = V3DTextureLoadFromFile2D(
                        tmp_path,
                        "mission_map_tex",	/* Not used. */
                        V3D_TEX_FORMAT_RGB,	/* Destination format. */
                        NULL,			/* Client data. */
                        NULL			/* Progress notify cb. */
                    );
                }
                StringFreeArray(strv, strc);

            }
        }


        /* Seek through all `create_helipad' parameters to add helipad
	 * markings to the map.
	 */
	rewind(fp);
	strptr = NULL;
	marking_ptr = NULL;
	while(1)
	{
	    /* Seek to next parameter. */
	    strptr = FSeekNextParm(fp, strptr, SAR_COMMENT_CHAR, '\0');
	    if(strptr == NULL)
		break;

	    /* Is it a create helipad parameter? */
	    if(!strcasecmp(strptr, "create_helipad") ||
               !strcasecmp(strptr, "new_helipad") ||
               !strcasecmp(strptr, "add_helipad")
	    )
	    {
		FSeekNextLine(fp);

                /* Allocate new marking on map for intercept. */
                marking_ptr = (sar_menu_map_marking_struct *)calloc(
                    1,
                    sizeof(sar_menu_map_marking_struct)
                );
                if(marking_ptr != NULL)
                {
                    marking_ptr->type = SAR_MENU_MAP_MARKING_TYPE_ICON;
                    marking_ptr->fg_color.a = 1.0;
                    marking_ptr->fg_color.r = 0.0;
                    marking_ptr->fg_color.g = 1.0;
                    marking_ptr->fg_color.b = 0.0;
                    marking_ptr->x = 0;
                    marking_ptr->y = 0;
                    marking_ptr->icon = core_ptr->menumap_helipad_img;

		    n = map_ptr->total_markings;
		    map_ptr->total_markings++;
		    map_ptr->marking = (sar_menu_map_marking_struct **)realloc(
			map_ptr->marking,
			map_ptr->total_markings * sizeof(sar_menu_map_marking_struct *)
		    );
		    if(map_ptr->marking == NULL)
		    {
			map_ptr->total_markings = 0;

			free(marking_ptr);
			marking_ptr = NULL;
		    }
		    else
		    {
			map_ptr->marking[n] = marking_ptr;
		    }
		}
	    }
	    /* Translation. */
            else if(!strcasecmp(strptr, "translate") ||
                    !strcasecmp(strptr, "translation")
	    )
            {
                FGetValuesF(fp, value, 3);
         
                if(marking_ptr != NULL)
                {
                    marking_ptr->x = value[0];
                    marking_ptr->y = value[1];

		    marking_ptr = NULL;
                }
            }
	    else
	    {
		FSeekNextLine(fp);
	    }
	}

	/* Close scene file. */
        FClose(fp);

        return(0);
}


/*
 *	Returns a pointer to a dynamically allocated description fetched
 *	from the specified mission file.
 */
char *SARMissionLoadDescription(const char *filename)
{
	int i, c;
	FILE *fp;
	char *parm, *desc_ptr = NULL;


	if(filename == NULL)
            return(desc_ptr);

	/* Open mission file. */
        fp = FOpen(filename, "rb");
        if(fp == NULL)
            return(desc_ptr);

	/* Check if an ASCII file. */
	for(i = 0; i < 256; i++)
	{
	    c = fgetc(fp);
	    if(c == EOF)
		break;

	    if(!isascii(c))
		return(desc_ptr);
	}

	/* Begin reading mission file. */
	rewind(fp);
        while(1)
        {
            c = fgetc(fp);
            if(c == EOF)
                break;
            
            /* Read past leading spaces. */
            if(ISBLANK(c))
            {
                FSeekPastSpaces(fp);
            
                /* Get first non space character. */
                c = fgetc(fp);
                if(c == EOF)
                    break;
            }

            /* Newline character? */
            if(ISCR(c))
                continue;

            /* Seek back one character, then fetch parameter. */
            fseek(fp, -1, SEEK_CUR);

            /* Get pointer to parameter string. */
            parm = SARMissionFSeekGetOpString(fp);

            /* Mission description? */
            if(!strcasecmp(parm, "description") ||
               !strcasecmp(parm, "desc")
            )
            {
		free(desc_ptr);
                desc_ptr = FGetString(fp);

		break;	/* Got what we wanted. */
            }
            /* Other. */
            else
            {
                FSeekNextLine(fp);
            }
	}

	/* Close mission file. */
	FClose(fp);

	return(desc_ptr);
}

/*
 *	Procedure to load `mission map' data from file and put into
 *	the menu's map object.
 */
void SARMissionLoadMenuMap(
	sar_core_struct *core_ptr,
	sar_menu_struct *m,
	const char *filename		/* Mission file name. */
)
{
	int i, n, c, map_num = -1;
	void *ptr;
	sar_menu_map_struct *map_ptr = NULL;
	FILE *fp;
	char *strptr, *scene_filename = NULL;
	char *arrive_at_name = NULL, *begin_at_name = NULL;
	double value[10];
	sar_menu_map_marking_struct	*marking_ptr = NULL,
					*intercept_marking_ptr = NULL;
	int intercepts_set = 0;


	if((core_ptr == NULL) ||
           (m == NULL)
	)
	    return;

	/* Get first map object on the menu. */
	for(i = 0; i < m->total_objects; i++)
	{
	    ptr = m->object[i];
	    if(ptr == NULL)
		continue;

	    if((*(int *)ptr) == SAR_MENU_OBJECT_TYPE_MAP)
	    {
		map_num = i;
		map_ptr = ptr;
		break;
	    }
	}
	if(map_ptr == NULL)
	    return;

	/* Delete all markings on menu's map object. */
	SARMenuMapDeleteAllMarkings(map_ptr);

	/* Note, order of which parameters appear in the mission file
	 * is important! We expect to parse certain parameters before
	 * others in this function.
	 */

	/* Open mission file. */
	fp = FOpen(filename, "rb");
	if(fp == NULL)
	    return;

        /* Begin reading mission file. */
        while(1)
        {
            c = fgetc(fp);
            if(c == EOF) 
                break;

            /* Read past leading spaces. */
            if(ISBLANK(c))
            {
                FSeekPastSpaces(fp);
  
                /* Get first non space character. */
                c = fgetc(fp);
                if(c == EOF)
                    break;
            }

            /* Newline character? */
            if(ISCR(c))
                continue;

            /* Comment? */
            if(ISCOMMENT(c))
            {
                FSeekNextLine(fp);	/* Seek to next line. */
                continue;
            }

            /* Seek back one character and fetch as operation. */
            fseek(fp, -1, SEEK_CUR);

            /* Get pointer to operation string. */
            strptr = SARMissionFSeekGetOpString(fp);

            /* Scene file. */
            if(!strcasecmp(strptr, "scene_file"))
            {
                strptr = FGetString(fp);   
                if(strptr != NULL)
                {
		    char tmp_path[PATH_MAX + NAME_MAX];

                    /* Complete path. */
                    if(ISPATHABSOLUTE(strptr))
                    {
                        strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                    }
                    else
                    {
			char *strptr2;
			struct stat stat_buf;

			strptr2 = PrefixPaths(dname.local_data, strptr);
                        if(strptr2 != NULL)
                        {  
                            if(stat(strptr2, &stat_buf))
                                strptr2 = PrefixPaths(
                                    dname.global_data,
                                    strptr
                                );
                        }
                        strncpy(
                            tmp_path,
                            ((strptr2 == NULL) ? strptr : strptr2),
                            PATH_MAX + NAME_MAX
                        );
                    }   
                    free(strptr);
		    strptr = NULL;

		    /* Update scene file name. */
		    free(scene_filename);
		    scene_filename = strdup(tmp_path);

		    /* Load information from scene file which are
		     * relivent to the mission map, such as map background
		     * texture file and landmarks.
		     */                 
		    SARMissionLoadMenpMapSceneInfo(
			core_ptr, m, map_ptr, scene_filename
		    );
                }
	    }
            /* Begin at name. */
            else if(!strcasecmp(strptr, "begin_at_name"))
            {
		sar_obj_intercept_struct intercept;

                memset(&intercept, 0x00, sizeof(sar_obj_intercept_struct));

		/* Update begin at name. */
		free(begin_at_name);
		begin_at_name = FGetString(fp);

		/* Always set first intercept to begin at name. */
                if(SARMissionGetInterceptNameFromScene(
		    scene_filename,
		    begin_at_name,
		    &intercept
		))
		    continue;

                /* Allocate new marking on map for intercept. */
                marking_ptr = (sar_menu_map_marking_struct *)calloc(
		    1,
		    sizeof(sar_menu_map_marking_struct)
		);
                intercept_marking_ptr = marking_ptr;
                if(marking_ptr != NULL)
                {
                    marking_ptr->type = SAR_MENU_MAP_MARKING_TYPE_INTERCEPT_LINE;
                    marking_ptr->fg_color.a = 1.0;
                    marking_ptr->fg_color.r = 0.0;   
                    marking_ptr->fg_color.g = 1.0;
                    marking_ptr->fg_color.b = 0.0;
                    marking_ptr->x = intercept.x;
                    marking_ptr->y = intercept.y;

		    intercepts_set++;

                    n = map_ptr->total_markings;
                    map_ptr->total_markings++;
                    map_ptr->marking = (sar_menu_map_marking_struct **)realloc(
			map_ptr->marking,
			map_ptr->total_markings * sizeof(sar_menu_map_marking_struct *)
		    );
		    if(map_ptr->marking == NULL)
                    {
                        map_ptr->total_markings = 0;
                        free(marking_ptr);   
                        marking_ptr = NULL;
                        intercept_marking_ptr = NULL;
                    }
                    else
                    {
                        map_ptr->marking[n] = marking_ptr;
                    }

		    /* Set initial scroll position on map with respect
		     * to the begin at location.
		     */
		    map_ptr->scroll_x = (int)(-intercept.x *
			map_ptr->m_to_pixels_coeff);
		    map_ptr->scroll_y = (int)(-intercept.y *
			map_ptr->m_to_pixels_coeff);

		    /* Reset meters to pixels coeff zoom. */
		    map_ptr->m_to_pixels_coeff = SAR_MAP_DEF_MTOP_COEFF;
                }
	    }
            /* Arrive at name. */
            else if(!strcasecmp(strptr, "arrive_at_name"))
            {
                free(arrive_at_name);
                arrive_at_name = FGetString(fp);
            }
            /* Add intercept. */
            else if(!strcasecmp(strptr, "add_intercept"))
            {
                sar_obj_intercept_struct intercept;

		/* Get intercept structure from file. */
                FGetValuesF(fp, value, 5);
                memset(&intercept, 0x00, sizeof(sar_obj_intercept_struct));

                switch((int)value[0])
                {
                  case 3:	/* Intercept arrive at location. */
		    SARMissionGetInterceptNameFromScene(
			scene_filename,
			arrive_at_name,
			&intercept
		    );
                    break;

                  case 2:	/* Intercept begin at location. */
                    SARMissionGetInterceptNameFromScene(
                        scene_filename,
                        begin_at_name,
                        &intercept
                    );
                    break;

		  default:	/* Standard intercept. */
                    intercept.x = value[1];
                    intercept.y = value[2];
                    intercept.z = SFMFeetToMeters(value[3]);
                    intercept.radius = value[4];
                    break;
                }

		/* Allocate new map marking structure, note previous
		 * marking structure should already be allocated from
		 * another add intercept parameter or begin at name
		 * parameter.
		 */
		if((marking_ptr != NULL) &&
		   (intercept_marking_ptr != NULL)
		)
		{
		    /* Allocate new intercept marking structure and
		     * copy over end position values from previous
		     * marking structure.
		     */
		    sar_menu_map_marking_struct *new_ptr =
			(sar_menu_map_marking_struct *)calloc(
                            1,
                            sizeof(sar_menu_map_marking_struct)
                    );

		    /* Set end position on last marking structure. */
                    marking_ptr->x_end = intercept.x;
                    marking_ptr->y_end = intercept.y;

		    intercepts_set++;

		    /* New marking structure allocated successfully? */
		    if(new_ptr != NULL)
		    {
			new_ptr->type = SAR_MENU_MAP_MARKING_TYPE_INTERCEPT_LINE;
			/* Copy over fg color from previous marking. */
			memcpy(
			    &new_ptr->fg_color,
			    &marking_ptr->fg_color,
			    sizeof(sar_menu_color_struct)
			);
			new_ptr->x = intercept.x;
			new_ptr->y = intercept.y;

			intercepts_set++;

			n = map_ptr->total_markings;
			map_ptr->total_markings++;
			map_ptr->marking = (sar_menu_map_marking_struct **)realloc(
			    map_ptr->marking,
			    map_ptr->total_markings * sizeof(sar_menu_map_marking_struct *)
			);
			if(map_ptr->marking == NULL)
			{
			    map_ptr->total_markings = 0;

			    free(new_ptr);

			    marking_ptr = NULL;
			    intercept_marking_ptr = NULL;
			}
			else
			{
			    map_ptr->marking[n] = new_ptr;
			}

			intercept_marking_ptr = new_ptr;
			marking_ptr = new_ptr;
		    }
		    else
		    {
			intercept_marking_ptr = NULL;
			marking_ptr = NULL;
		    }


		    /* Allocate another marking structure to represent
		     * way point icon.
		     */
		    new_ptr = (sar_menu_map_marking_struct *)calloc(
                        1,
                        sizeof(sar_menu_map_marking_struct)
                    );
		    if(new_ptr != NULL)
		    {
                        new_ptr->type = SAR_MENU_MAP_MARKING_TYPE_ICON;
                        new_ptr->x = intercept.x;
                        new_ptr->y = intercept.y;
                        new_ptr->icon = core_ptr->menumap_intercept_img;

                        n = map_ptr->total_markings;
                        map_ptr->total_markings++;
                        map_ptr->marking = (sar_menu_map_marking_struct **)realloc(
                            map_ptr->marking,
			    map_ptr->total_markings * sizeof(sar_menu_map_marking_struct *)
                        );
                        if(map_ptr->marking == NULL)
                        {
                            map_ptr->total_markings = 0;

                            free(new_ptr);
                        }
                        else
                        {
                            map_ptr->marking[n] = new_ptr;
                        }
		    }

                }
	    }
            /* Skip other parameters. */
            else
            {
                FSeekNextLine(fp);	/* Seek to next line. */
	    }
	}

	/* Close file. */
	FClose(fp);


	/* Deallocate loaded names. */
	free(scene_filename);
	scene_filename = NULL;

	free(arrive_at_name);
	arrive_at_name = NULL;

	free(begin_at_name);
	begin_at_name = NULL;


	/* If the number of intercepts set on the map is an odd number,
	 * then that implies that the last intercept line marking is
	 * extraneous.
	 */
	if(intercepts_set & 1)
	{
	    /* Delete the last marking pointer of type
	     * SAR_MENU_MAP_MARKING_TYPE_INTERCEPT_LINE. Itterate from
	     * last to first of all markings on the map.
	     */
	    for(i = map_ptr->total_markings - 1; i >= 0; i--)
	    {
		marking_ptr = map_ptr->marking[i];
		if(marking_ptr == NULL)
		    continue;

		/* Check marking type. */
		if(marking_ptr->type == SAR_MENU_MAP_MARKING_TYPE_INTERCEPT_LINE)
		{
		    /* Need to delete this marking on the map, set the
		     * pointer to NULL after deleting it.
		     */
		    free(marking_ptr);
		    map_ptr->marking[i] = NULL;

		    /* That's it, don't delete any more markings. */
		    break;
		}
	    }
	}

	return;
}


/*
 *	Loads a new mission
 *
 *	Note that the scene structure in the core structure must be
 *	allocated and with all values reset.
 */
sar_mission_struct *SARMissionLoadFromFile(
	sar_core_struct *core_ptr,
	const char *filename,
	void *client_data,
	int (*progress_func)(void *, long, long)
)
{
	int c, status, lines_read = 0;
	int loaded_scene = 0;
        FILE *fp;
        char *strptr, *strptr2;
	long file_size;

	sar_scene_struct *scene;
	sar_mission_struct *mission;
	sar_mobj_arrive_struct *arrive_objective = NULL;
	sar_mobj_rescue_struct *rescue_objective = NULL;

	int humans_need_rescue = 0;
	char *begin_at_name = NULL, *arrive_at_name = NULL;

	char *weather_preset_name = NULL;

	int obj_num = -1;
	sar_object_struct *obj_ptr = NULL;
	sar_position_struct *pos;
	sar_position_struct begin_pos;
	sar_direction_struct begin_dir;

        double value[10];
        struct stat stat_buf;
	char tmp_path[PATH_MAX + NAME_MAX];


        if((core_ptr == NULL) ||
           (filename == NULL)
        )
            return(NULL);

	/* Get pointer to scene structure. */
	scene = core_ptr->scene;
	if(scene == NULL)
	    return(NULL);

	/* Check if mission file exists. */
        if(stat(filename, &stat_buf))
            return(NULL);
        file_size = stat_buf.st_size;

        /* Note, order of which parameters appear in the mission file
         * is important! We expect to parse certain parameters before
         * others in this function.
         */

	/* Open mission file. */
        fp = FOpen(filename, "rb");
        if(fp == NULL)
            return(NULL);

	/* Allocate a new mission structure. */
	mission = (sar_mission_struct *)calloc(
	    1,
	    sizeof(sar_mission_struct)
	);
	if(mission == NULL)
	{
	    FClose(fp);
	    return(NULL);
	}


	/* Set up default mission values. */
	mission->type = MISSION_TYPE_NONE;
	mission->state = MISSION_STATE_IN_PROGRESS;
	mission->title = NULL;
	mission->time_spent = 0.0;
	mission->check_int = 1000;		/* In milliseconds. */
	mission->next_check = cur_millitime + mission->check_int;


        /* Begin reading mission file. */
        while(1)
        {
            c = fgetc(fp);
            if(c == EOF) 
                break;

            /* Read past leading spaces. */
            if(ISBLANK(c))
            {
                FSeekPastSpaces(fp);

                /* Get first non space character. */
                c = fgetc(fp);
                if(c == EOF)
                    break;
            }

            /* Newline character? */
            if(ISCR(c))
                continue;

	    /* Comment? */
            if(ISCOMMENT(c))
            {
                FSeekNextLine(fp);	/* Seek to next line. */
                continue;
            }

            /* Seek back one character and fetch as operation. */
            fseek(fp, -1, SEEK_CUR);

            /* Get pointer to operation string. */
            strptr = SARMissionFSeekGetOpString(fp);

            /* Mission name (title). */
            if(!strcasecmp(strptr, "name"))
            {
		strptr = FGetString(fp);

		/* Set new mission title. */
		free(mission->title);
		mission->title = strptr;
            }
	    /* Mission description. */
	    else if(!strcasecmp(strptr, "description") ||
                    !strcasecmp(strptr, "desc")
	    )
            {
		strptr = FGetString(fp);

		/* Set new mission description. Note that FGetString()
		 * will parse escape sequences, like escaped newlines
		 * will be throw out and escaped 'n' characters will be
		 * saved as newline characters.
		 */
                free(mission->description);
                mission->description = strptr;
            }
	    /* Mission type. */
	    else if(!strcasecmp(strptr, "type"))
            {
		int mission_type;

		FGetValuesF(fp, value, 1);
		mission_type = (int)value[0];

		/* Allocate a new mission objective structure
		 * depending on the type of mission.
		 */
		switch(mission_type)
		{
		  case MISSION_TYPE_NONE:
		    mission->type = mission_type;
		    mission->objective = NULL;
		    break;

		  case MISSION_TYPE_ARRIVE:
		    arrive_objective = (sar_mobj_arrive_struct *)calloc(
			1,
			sizeof(sar_mobj_arrive_struct)
		    );
		    if(arrive_objective == NULL)
			break;

		    mission->type = mission_type;
		    mission->objective = arrive_objective;

		    arrive_objective->type = mission_type;
		    arrive_objective->time_left = 0.0;
		    arrive_objective->arrive_at_name = NULL;
		    arrive_objective->message_success = NULL;
		    arrive_objective->message_fail = NULL;
		    break;

                  case MISSION_TYPE_RESCUE:
                    rescue_objective = (sar_mobj_rescue_struct *)calloc(
                        1,
                        sizeof(sar_mobj_rescue_struct)
                    );
                    if(rescue_objective == NULL)
                        break;

                    mission->type = mission_type;
                    mission->objective = rescue_objective;

		    rescue_objective->type = mission_type;
		    rescue_objective->time_left_type = MISSION_TIME_LEFT_TYPE_RESCUE_ALL;
                    rescue_objective->time_left = 0.0;
		    rescue_objective->humans_rescued = 0;
                    rescue_objective->humans_need_rescue = 0;
                    rescue_objective->humans_died = 0;
                    rescue_objective->message_success = NULL;
                    rescue_objective->message_fail = NULL;
                    break;

		  default:
		    fprintf(
			stderr,
			"%s: Line %i: Unsupported mission type `%i'\n",
			filename, lines_read, mission_type
		    );
		    break;
		}
            }
            /* Mission map. */
            else if(!strcasecmp(strptr, "mission_map"))
            {
                FSeekNextLine(fp);
	    }
	    /* Scene file. */
            else if(!strcasecmp(strptr, "scene_file"))
            {
		if(loaded_scene)
		{
		    fprintf(
			stderr,
			"%s: Line %i: Redefination of scene_file ignored.\n",
                        filename, lines_read
                    );
		    FSeekNextLine(fp);
		}
		else
		{
		    strptr = FGetString(fp);

		    if(strptr != NULL)
		    {
                        /* Complete path. */
                        if(ISPATHABSOLUTE(strptr))
                        {
                            strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                        }
                        else
                        {   
                            strptr2 = PrefixPaths(dname.local_data, strptr);
                            if(strptr2 != NULL)
                            {
                                if(stat(strptr2, &stat_buf))
                                    strptr2 = PrefixPaths(
					dname.global_data,
					strptr
				    );
			    }
			    strncpy(
                                tmp_path,
                                ((strptr2 == NULL) ? strptr : strptr2),
                                PATH_MAX + NAME_MAX
                            );
                        }
		        free(strptr);

			/* Set new scene file on mission structure. */
                        free(mission->scene_file);
                        mission->scene_file = StringCopyAlloc(tmp_path);

			/* Load scene. */
			status = SARSceneLoadFromFile(
			    core_ptr,
			    core_ptr->scene,
			    mission->scene_file,
			    weather_preset_name,
			    core_ptr,
			    progress_func
			);
			if(status)
			{
			    /* Failed to load scene. */
			    fprintf(
                                stderr,
               "%s: Line %i: Error occured while loading scene %s.\n",
				filename, lines_read, mission->scene_file
			    );

			    SARSceneDestroy(
				core_ptr,
				core_ptr->scene,
				&core_ptr->object,
				&core_ptr->total_objects
			    );

			    SARMissionDestroy(mission);
			    mission = NULL;

			    break;
			}
			else
			{
			    loaded_scene = 1;
			}
		    }
		}
	    }
            /* Player model file. */
            else if(!strcasecmp(strptr, "player_model_file"))
            {
                strptr = FGetString(fp);

                if((strptr != NULL) && loaded_scene)
                {
		    /* Complete path. */
		    if(ISPATHABSOLUTE(strptr))
		    {
			strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
		    }
		    else
		    {
			strptr2 = PrefixPaths(dname.local_data, strptr);
			if(strptr2 != NULL)
			{
			    if(stat(strptr2, &stat_buf))
				strptr2 = PrefixPaths(dname.global_data, strptr);
			}
			strncpy(
			    tmp_path,
			    ((strptr2 == NULL) ? strptr : strptr2),
			    PATH_MAX + NAME_MAX
			);
		    }
		    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
		    free(strptr);
		    strptr = NULL;

		    /* Update player model file name on mission and
		     * core structure.
		     */
                    free(mission->player_model_file);
                    mission->player_model_file = strdup(tmp_path);

                    free(core_ptr->cur_player_model_file);
                    core_ptr->cur_player_model_file = strdup(tmp_path);

		    /* Create player object. */
		    SARSceneAddPlayerObject(
			core_ptr, scene,
			tmp_path,
			&begin_pos, NULL
		    );

		    /* Update current object context. */
		    obj_num = scene->player_obj_num;
		    obj_ptr = scene->player_obj_ptr;
                }
            }
            /* Player stats file. */
            else if(!strcasecmp(strptr, "player_stats_file"))
            {
                strptr = FGetString(fp);

                if(strptr != NULL)
                {
                    /* Complete path. */   
                    if(ISPATHABSOLUTE(strptr))
                    { 
                        strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                    }
                    else
                    {
                        strptr2 = PrefixPaths(dname.local_data, strptr);
                        if(strptr2 != NULL)
                        {
                            if(stat(strptr2, &stat_buf))
                                strptr2 = PrefixPaths(dname.global_data, strptr);
                        }
                        strncpy(
                            tmp_path,
                            ((strptr2 == NULL) ? strptr : strptr2),
                            PATH_MAX + NAME_MAX
                        );
                    }
                    tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';
                    free(strptr);

                    free(mission->player_stats_file);
                    mission->player_stats_file = strdup(tmp_path);
                }
            }

	    /* Time left. */
	    else if(!strcasecmp(strptr, "time_left"))
            {
		int time_left_type;
		double time_left;

		/* Arguments: <time_left_type> <time_left_sec> */
                FGetValuesF(fp, value, 2);
		time_left_type = (int)value[0];
		time_left = value[1];

		if(arrive_objective != NULL)
		{
		    /* Ignore time_left_type for arrive objectives. */
		    arrive_objective->time_left = time_left;
		}
		else if(rescue_objective != NULL)
		{
		    rescue_objective->time_left_type = time_left_type;
		    rescue_objective->time_left = time_left;
		}
	    }
	    /* Begin at name. */
            else if(!strcasecmp(strptr, "begin_at_name"))
            {
		sar_object_struct *obj_ptr2;

		/* Update mission begin at name. */
		free(begin_at_name);
                begin_at_name = FGetString(fp);

		/* Match begin at object. */
		obj_ptr2 = SARObjMatchPointerByName(
                    scene, core_ptr->object, core_ptr->total_objects,
                    begin_at_name, NULL
                );
		/* Record begin at position and direction. */
		if(obj_ptr2 != NULL)
		{
		    memcpy(&begin_pos, &obj_ptr2->pos, sizeof(sar_position_struct));
/* Need to work on this. */
		}
	    }
            /* Arrive at name. */
            else if(!strcasecmp(strptr, "arrive_at_name"))
            {
		char **storage = NULL;

		free(arrive_at_name);
                arrive_at_name = FGetString(fp);

                if(arrive_objective != NULL)
		    storage = &arrive_objective->arrive_at_name;
                else if(rescue_objective != NULL)
		    storage = &rescue_objective->arrive_at_name;

		if(storage != NULL)
		{
		    free(*storage);
		    *storage = StringCopyAlloc(arrive_at_name);
		}
            }
            /* Message success. */
            else if(!strcasecmp(strptr, "message_success"))
            {
                char **storage = NULL;

                strptr = FGetString(fp);

                if(arrive_objective != NULL)
                    storage = &arrive_objective->message_success;
                else if(rescue_objective != NULL)
                    storage = &rescue_objective->message_success;

                if(storage != NULL)
                {
                    free(*storage);
                    *storage = strptr;
                }
                else
                {
                    free(strptr);
                }
            }
            /* Message fail. */
            else if(!strcasecmp(strptr, "message_fail"))
            {
                char **storage = NULL;

                strptr = FGetString(fp);

                if(arrive_objective != NULL)
                    storage = &arrive_objective->message_fail;
                else if(rescue_objective != NULL)
                    storage = &rescue_objective->message_fail;

                if(storage != NULL)
                {
                    free(*storage);   
                    *storage = strptr;
                }   
                else
                {
                    free(strptr);
                }
            }    

	    /* Weather. */
	    else if(!strcasecmp(strptr, "weather"))
            {
                strptr = FGetString(fp);

		if(loaded_scene)
		{
                    fprintf(stderr,
 "%s: Line %i: Cannot specify `weather' after loading scene.\n",
                        filename, lines_read
                    );
		    free(strptr);
		}
		else
		{
		    free(weather_preset_name);
		    weather_preset_name = strptr;
		}
	    }
	    /* Time of day. */
            else if(!strcasecmp(strptr, "time_of_day"))
            {
                strptr = FGetString(fp);

                if(loaded_scene)
                {
/* Parse and set time of day. */
/*		    scene->tod = ???; */
		}
                else
                {
                    fprintf(stderr,
 "%s: Line %i: Cannot specify `time_of_day' before loading scene.\n",
                        filename, lines_read
                    );
                }

                free(strptr);
	    }
            /* Total number of humans that need rescue. */
            else if(!strcasecmp(strptr, "humans_need_rescue"))
            {
                FGetValuesF(fp, value, 1);

		if(rescue_objective != NULL)
		{
		    rescue_objective->humans_need_rescue = (int)value[0];

		    /* Check if the number of humans that need rescue
		     * that were actually added are fewer than
		     * this amount (this parameter should be specified
		     * after all the humans that need rescue have been
		     * added).
		     */
		    if(humans_need_rescue <
			rescue_objective->humans_need_rescue
		    )
			fprintf(stderr,
 "%s: Line %i: Warning: humans_need_rescue %i less than actual %i.\n",
                            filename, lines_read,
			    rescue_objective->humans_need_rescue,
			    humans_need_rescue
			);
		}
            }

	    /* Select object by name. */
	    else if(!strcasecmp(strptr, "select_object_by_name"))
            {
                strptr = FGetString(fp);

		if(strptr != NULL)
		{
		    /* Update current object context. */
		    obj_ptr = SARObjMatchPointerByName(
			scene,
			core_ptr->object, core_ptr->total_objects,
			strptr, &obj_num
		    );

		    free(strptr);
		    strptr = NULL;
		}
	    }
            /* Translation. */
            else if(!strcasecmp(strptr, "translate") ||
                    !strcasecmp(strptr, "translation")
	    )
            {
		sar_position_struct new_pos;


                FGetValuesF(fp, value, 3);

                new_pos.x = value[0];
                new_pos.y = value[1];
                new_pos.z = SFMFeetToMeters(value[2]);

		if(obj_ptr != NULL)
		    SARSimWarpObject(scene, obj_ptr, &new_pos, NULL);
            }
            /* Rotate. */
            else if(!strcasecmp(strptr, "rotate"))
            {
                sar_direction_struct new_dir;


                FGetValuesF(fp, value, 3);

                /* Values are in degrees. */
                new_dir.heading = SFMDegreesToRadians(value[0]);
                new_dir.pitch = SFMDegreesToRadians(value[1]);
                new_dir.bank = SFMDegreesToRadians(value[2]);

                if(obj_ptr != NULL)
		    SARSimWarpObject(scene, obj_ptr, NULL, &new_dir);
            }
	    /* Randomize position (cylendrical coordinates). */
	    else if(!strcasecmp(strptr, "randomize_translation_cylendrical") ||
		    !strcasecmp(strptr, "randomize_translation_cyl") ||
		    !strcasecmp(strptr, "rand_translation_cyl")
	    )
	    {
		/* Arguments: <r_rand> <z_rand> */
                FGetValuesF(fp, value, 2);

                /* Values are in degrees. */
                if(obj_ptr != NULL)
                {
		    int rand_val = SARRandom();
		    double r = value[0], z = value[1];
		    double rand_theta, rand_coeff;

		    pos = &obj_ptr->pos;
		    rand_theta = SFMDegreesToRadians((double)(rand_val % 360));
		    rand_coeff = (double)rand_val / (double)RAND_MAX;

		    pos->x += (sin(rand_theta) * rand_coeff * r);
		    pos->y += (cos(rand_theta) * rand_coeff * r);
		    pos->z += (rand_coeff * z);
		}
	    }
	    /* Add object. */
            else if(!strcasecmp(strptr, "add_object"))
            {
                strptr = FGetString(fp);
 		if(strptr != NULL)
		{
		    int strc;
		    char **strv;

		    strv = strexp(strptr, &strc);

		    /* Format: <type> <range> <model_file> */
                    obj_num = SARObjCreate(
                        scene, &core_ptr->object, &core_ptr->total_objects,
			((strc > 0) ? atoi(strv[0]) : SAR_OBJ_TYPE_STATIC)
		    );
                    obj_ptr = ((obj_num < 0) ? NULL : core_ptr->object[obj_num]);

		    if((obj_ptr != NULL) && (strc > 2))
		    {
			SARObjLoadFromFile(
			    core_ptr,
			    obj_num,
			    strv[2]
			);

			obj_ptr->range = atof(strv[1]);
		    }

		    StringFreeArray(strv, strc);
		    free(strptr);
		    strptr = NULL;
		}
	    }
	    /* Add premodeled object. */
            else if(!strcasecmp(strptr, "add_premodeled_object"))
            {
                strptr = FGetString(fp);
                if(strptr != NULL)
                {
                    int strc;
                    char **strv;

                    strv = strexp(strptr, &strc);

                    obj_num = SARObjCreatePremodeled(
                        core_ptr, scene, strc, strv
                    );
                    obj_ptr = ((obj_num < 0) ?
			NULL : core_ptr->object[obj_num]
		    );

                    StringFreeArray(strv, strc);
                }
                free(strptr);
		strptr = NULL;
	    }
            /* New human object. */
            else if(!strcasecmp(strptr, "create_human") ||
                    !strcasecmp(strptr, "new_human") ||
                    !strcasecmp(strptr, "add_human")
            )
            {
                strptr2 = FGetString(fp);
                obj_num = SARObjLoadLineHuman(
                    core_ptr, scene, strptr2
                );
                obj_ptr = ((obj_num < 0) ?
		    NULL : core_ptr->object[obj_num]
		);
                free(strptr2);
		strptr2 = NULL;

		/* Increment number of humans that need rescue that
		 * this mission file has specified.
		 */
		humans_need_rescue++;

		/* Update total on rescue objective structure if
		 * this is a rescue mission.
		 */
		if(rescue_objective != NULL)
		{
		    rescue_objective->humans_need_rescue =
			humans_need_rescue;
		}
            }
	    /* Set human rescue message. */
            else if(!strcasecmp(strptr, "set_human_message_hoist_in"))
            {
		strptr = FGetString(fp);
                if((strptr != NULL) && (obj_ptr != NULL))
		{
                    if(obj_ptr->type == SAR_OBJ_TYPE_HUMAN)
                    {
                        sar_object_human_struct *obj_human_ptr = obj_ptr->data;

			if(obj_human_ptr != NULL)
			{
			    free(obj_human_ptr->mesg_hoist_in);
			    obj_human_ptr->mesg_hoist_in = StringCopyAlloc(strptr);
			}
                    }
		    else
		    {
			fprintf(
			    stderr,
 "%s: Line %i: set_human_message_hoist_in: Object not of type %i.\n",
			    filename, lines_read, SAR_OBJ_TYPE_HUMAN
			);
		    }
                }
		free(strptr);
		strptr = NULL;
	    }

	    /* Add intercept to current object in context. */
            else if(!strcasecmp(strptr, "add_intercept"))
            {
		sar_obj_intercept_struct intercept;
		sar_object_struct *tar_obj_ptr;

                FGetValuesF(fp, value, 6);
		memset(&intercept, 0x00, sizeof(sar_obj_intercept_struct));

		/* Argument is handled by first checking the first
		 * argument. If the first argument is 2 or 3 then
		 * the argument is handled as take values from the
		 * begin at or arrive at object (respectivly).
		 * The arguments would be as follows:
		 *
		 *	<2|3> <radius> <urgency>
		 *
		 * If first argument is 1 then the arguments would then
		 * be handled as follows:
		 *
		 *	1 <x> <y> <z> <radius> <urgency>
		 */
		if(obj_ptr != NULL)
		{
		    sar_object_aircraft_struct *obj_aircraft_ptr;

		    /* Handle by intercept referance. */
		    switch((int)value[0])
		    {
		      case 3:	/* Intercept arrive at location. */
                        tar_obj_ptr = SARObjMatchPointerByName(
                            scene,
                            core_ptr->object, core_ptr->total_objects,
                            arrive_at_name, NULL
                        );
                        if(tar_obj_ptr != NULL)
			{
			    intercept.x = tar_obj_ptr->pos.x;
			    intercept.y = tar_obj_ptr->pos.y;
                            intercept.z = tar_obj_ptr->pos.z;
			    intercept.radius = value[1];
			    intercept.urgency = value[2];
			}
			break;

                      case 2:	/* Intercept begin at location. */
                        tar_obj_ptr = SARObjMatchPointerByName(
                            scene,
                            core_ptr->object, core_ptr->total_objects,
                            begin_at_name, NULL
                        );
                        if(tar_obj_ptr != NULL)
                        {
                            intercept.x = tar_obj_ptr->pos.x;
                            intercept.y = tar_obj_ptr->pos.y;
                            intercept.z = tar_obj_ptr->pos.z;
                            intercept.radius = value[1];
			    intercept.urgency = value[2];
                        }
                        break;

		      default:	/* Standard intercept. */
			intercept.x = value[1];
			intercept.y = value[2];
			intercept.z = SFMFeetToMeters(value[3]);
			intercept.radius = value[4];
			intercept.urgency = value[5];
			break;
		    }

		    /* Allocate new intercept structure on object. */
		    switch(obj_ptr->type)
		    {
		      case SAR_OBJ_TYPE_AIRCRAFT:
			obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
			if(obj_aircraft_ptr != NULL)
			{
			     SARObjCreateIntercept(
				scene,
				&obj_aircraft_ptr->intercept,
				&obj_aircraft_ptr->total_intercepts,
				intercept.flags,
				intercept.x, intercept.y, intercept.z,
				intercept.radius, intercept.urgency,
				intercept.name
			    );
			}
			break;
		    }

		}
	    }

            /* Unknown operation. */
            else 
            {       
                fprintf(
                    stderr,
                    "%s: Line %i: Unsupported operation `%s'\n",
                    filename, lines_read, strptr
                );
                FSeekNextLine(fp);	/* Seek to next line. */
            }
	}

	/* Close mission file. */
	FClose(fp);


	/* ********************************************************* */
	/* Post loading checks (warning, mission might be NULL). */

	/* Arrive at object exists? */
	if(arrive_objective != NULL)
	{
	    strptr = arrive_objective->arrive_at_name;
	    if(strptr != NULL)
	    {
                obj_ptr = SARObjMatchPointerByName(
                    scene,
                    core_ptr->object, core_ptr->total_objects,
		    strptr, NULL
		);
		if(obj_ptr == NULL)
		{
		    fprintf(
			stderr,
 "%s: Warning: No object in scene matching arrive_at_name `%s'\n",
			filename,
			strptr
		    );
		}
	    }
	}
	/* Rescue objective. */
        else if(rescue_objective != NULL)
        {
            strptr = rescue_objective->arrive_at_name;
            if(strptr != NULL)
            {
                obj_ptr = SARObjMatchPointerByName(
                    scene,
                    core_ptr->object, core_ptr->total_objects,
                    strptr, NULL                          
                );      
                if(obj_ptr == NULL)
                {
                    fprintf(
                        stderr,
 "%s: Warning: No object in scene matching arrive_at_name `%s'\n",
                        filename,
                        strptr                          
                    );
                }
            }    
	}


	/* Deallocate names. */
	free(begin_at_name);
	free(arrive_at_name);

	free(weather_preset_name);


	return(mission);
}
