#include <stdio.h>  
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>

#include "../include/os.h"

#ifdef __MSW__
# include <windows.h>
#else
# include <sys/time.h>
# include <unistd.h>
#endif

#include <GL/gl.h>
#include "../include/strexp.h"
#include "../include/string.h"

#include "menu.h"
#include "obj.h"

#include "sar.h"   


time_t SARGetCurMilliTime(void);

int SARRandom(void);

Boolean SARGetGLVersion(int *major, int *minor, int *release);
char *SARGetGLVendorName(void);
char *SARGetGLRendererName(void);
char **SARGelGLExtensionNames(int *strc);

int SARIsMenuAllocated(sar_core_struct *core_ptr, int n);
int SARMatchMenuByName(
        sar_core_struct *core_ptr,
        const char *name
);
sar_menu_struct *SARGetCurrentMenuPtr(sar_core_struct *core_ptr);
void SARDeleteListItemData( 
        sar_menu_list_item_struct *item_ptr
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *      Returns the current time since midnight in milliseconds from the
 *      system.
 */
time_t SARGetCurMilliTime(void)
{
#ifdef __MSW__
        SYSTEMTIME t;   /* Current time of day structure. */

        GetSystemTime(&t);
        return(
            (time_t)(
                (((((t.wHour * 60.0) + t.wMinute) * 60) + t.wSecond) * 1000) +
                t.wMilliseconds
            )
        );
#else
        struct timeval tv[1];

        if(gettimeofday(tv, NULL) < 0)
            return(-1);

        return(((tv->tv_sec % 86400) * 1000) + (tv->tv_usec / 1000));
#endif
}


/*
 *      Returns a random number from rand().
 */
int SARRandom(void)
{
        srand((unsigned int)cur_millitime);
        return(rand());
}


/*
 *	Returns the version numbers for the GL version, returns
 *	TRUE on success.
 *
 *	If any of the inputs are NULL, then that respective input will
 *	not be set.
 */
Boolean SARGetGLVersion(int *major, int *minor, int *release)
{
	const GLubyte *strptr;
	char **strv;
	int strc;


	/* Reset inputs. */
	if(major != NULL)
	    (*major) = 0;
        if(minor != NULL)
            (*minor) = 0;
        if(release != NULL)
            (*release) = 0;

	/* Get statically allocated '.' character separated string
	 * containing the version numbers. Format can be either
	 * "major.minor" or "major.minor.release".
	 */
	strptr = glGetString(GL_VERSION);
	if(strptr == NULL)
	    return(False);

	/* Explode version string at the '.' characters. */
	strv = strchrexp((const char *)strptr, '.', &strc);
	if(strv == NULL)
	    return(False);

	if((strc > 0) && (major != NULL))
	    (*major) = atoi(strv[0]);

        if((strc > 1) && (minor != NULL))
            (*minor) = atoi(strv[1]);

        if((strc > 2) && (release != NULL))
            (*release) = atoi(strv[2]);

	StringFreeArray(strv, strc);

	return(True);
}

/*
 *      Returns a dynamically allocated string containing the GL
 *      vendor name. The returned string must be free()'ed by the
 *      calling function.
 *
 *      Can return NULL on error.
 */
char *SARGetGLVendorName(void)
{
        const GLubyte *strptr;

        /* Get statically allocated string containing the GL renderer
         * name.
         */
        strptr = glGetString(GL_VENDOR);
        if(strptr == NULL)
            return(NULL);

        /* Return dynamically allocated string. */
        return(strdup((const char *)strptr));
}

/*
 *	Returns a dynamically allocated string containing the GL
 *	renderer name. The returned string must be free()'ed by the
 *	calling function.
 *
 *	Can return NULL on error.
 */
char *SARGetGLRendererName(void)
{
	const GLubyte *strptr;


	/* Get statically allocated string containing the GL renderer
	 * name.
	 */
	strptr = glGetString(GL_RENDERER);
	if(strptr == NULL)
	    return(NULL);

	/* Return dynamically allocated string. */
	return(strdup((const char *)strptr));
}


/*
 *	Returns a dynamically allocated array of strings containing
 *	the list of GL extension names available.
 *
 *	If strc is not NULL then it will be set to the number of
 *	strings returned.
 *
 *	Can return NULL on error.
 */
char **SARGelGLExtensionNames(int *strc)
{
        const GLubyte *strptr;
	char **strv;
	int lstrc;


	/* Reset inputs. */
	if(strc != NULL)
	    (*strc) = 0;

	/* Get statically allocated space separated string containing
	 * the GL extensions list.
	 */
        strptr = glGetString(GL_EXTENSIONS);
	if(strptr == NULL)
	    return(NULL);

	/* Explode space separated string. */
	strv = strexp(strptr, &lstrc);
	if(strv == NULL)
	    return(NULL);

	/* Update number of strings reciefved (hence number of
	 * extensions).
	 */
	if(strc != NULL)
	    (*strc) = lstrc;

	return(strv);
}


/*
 *      Is menu allocated on the core structure.
 */
int SARIsMenuAllocated(sar_core_struct *core_ptr, int n)
{
        if(core_ptr == NULL)
            return(0);
        else if((n < 0) || (n >= core_ptr->total_menus))
            return(0);
        else if(core_ptr->menu[n] == NULL)
            return(0);
        else
            return(1);
}

/*
 *      Matches a menu on the core structure that matches the given
 *      name (case insensitive). Can return -1 for no match.
 */
int SARMatchMenuByName(
        sar_core_struct *core_ptr,
        const char *name
)
{
        int i;
        sar_menu_struct *menu_ptr;
        
        
        if((core_ptr == NULL) ||
           (name == NULL)
        )
            return(-1);
            
        for(i = 0; i < core_ptr->total_menus; i++)
        {
            menu_ptr = core_ptr->menu[i];
            if(menu_ptr == NULL)
                continue;

            if(menu_ptr->name == NULL)
                continue;
            
            if(!strcasecmp(menu_ptr->name, name))
                return(i);
        }
 
        return(-1);
}

/*
 *      Get pointer to currently selected menu, can return NULL on
 *      error or if no menu is selected.
 */
sar_menu_struct *SARGetCurrentMenuPtr(sar_core_struct *core_ptr)
{
        int n;
        
        
        if(core_ptr == NULL)
            return(NULL);
        else
            n = core_ptr->cur_menu;
         
        if((n < 0) || (n >= core_ptr->total_menus))
            return(NULL);
        else
            return(core_ptr->menu[n]);
}

/*
 *      Deletes the client data structure specified on the menu
 *      list object item pointer.
 */
void SARDeleteListItemData(
        sar_menu_list_item_struct *item_ptr
)
{
        sar_menu_list_item_data_struct *li_data_ptr;

        if(item_ptr == NULL)
            return;
        else
            li_data_ptr = item_ptr->client_data;

        free(li_data_ptr->filename);
        free(li_data_ptr->name);

        free(li_data_ptr);

        item_ptr->client_data = NULL;

        return;
}
