#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>

#include "sfm.h"
#include "obj.h"

#include "simutils.h"
#include "simsurface.h"

#include "sar.h"


double SARSimSupportSurfaceHeight(
	sar_object_struct *obj_ptr,
	sar_contact_bounds_struct *cb,
	sar_position_struct *pos
);
double SARSimHFGetGroundHeight(
        sar_object_struct *obj_ptr,
        sar_position_struct *pos
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))


/*
 *	Checks the contact bounds cb (which should have come from
 *	obj_ptr) against the given pos.
 *
 *	Returns the z value of the obj_ptr's position plus any contact
 *	bound increased z offsets if and only if the given source pos
 *	is over the obj_ptr. Checks here take into consideration the
 *	hollowness of the target object, if the source position is below
 *	the contact bounds of the target object then the z value of the
 *	target object position is not returned.
 *
 *	Calls to this function should already check if cb's crash_flags
 *	specify SAR_CRASH_FLAG_SUPPORT_SURFACE before calling this
 *	function. So this function assumes the contact bounds specify
 *	the object to have a landable/walkable surface.
 */
double SARSimSupportSurfaceHeight(
	sar_object_struct *obj_ptr,	/* Target. */
        sar_contact_bounds_struct *cb,	/* Target. */
        sar_position_struct *pos	/* Source. */
)
{
	double dx, dy, dx2, dy2, dr, cr;
	double x_min, x_max, y_min, y_max;
	double cur_z_height = 0.0;
	sar_contact_bounds_struct *cb_tar;
	sar_position_struct *pos_src, *pos_tar;


        if((obj_ptr == NULL) ||
           (cb == NULL) ||
           (pos == NULL)
        )
            return(cur_z_height);

	cb_tar = cb;
	pos_tar = &obj_ptr->pos;
	pos_src = pos;

	/* Handle by target object contact bounds shape. */
	switch(cb_tar->contact_shape)
	{
	  case SAR_CONTACT_SHAPE_SPHERICAL:
            dx = pos_src->x - pos_tar->x;
            dy = pos_src->y - pos_tar->y;
            cr = cb_tar->contact_radius;
            dr = hypot(dx, dy);

            /* Delta distance apart equal or less than contact radius? */
            if(dr <= cr)
            {
		/* Use spherical radius as contact z height offset. */
                double check_z = pos_tar->z + cr;

                /* Is source position is under the target surface? */
                if(pos_src->z < check_z)
                {
                    /* Source object is under, so do not account for
                     * surface. This will make any source position under
                     * the target surface `hollow'.
                     */
                }
                else
                {
                    cur_z_height = check_z;
                }
            }
	    break;

	  case SAR_CONTACT_SHAPE_CYLENDRICAL:
	    dx = pos_src->x - pos_tar->x;
	    dy = pos_src->y - pos_tar->y;
	    cr = cb_tar->contact_radius;
	    dr = hypot(dx, dy);

	    /* Delta distance apart equal or less than contact radius? */
	    if(dr <= cr)
	    {
		double check_z = pos_tar->z + cb_tar->contact_h_max;

		/* Is source position is under the target surface? */
		if(pos_src->z < check_z)
		{
		    /* Source object is under, so do not account for
		     * surface. This will make any source position under
		     * the target surface `hollow'.
		     */
		}
		else
		{
		    cur_z_height = check_z;
		}
	    }
	    break;

	  case SAR_CONTACT_SHAPE_RECTANGULAR:
            dx = pos_src->x - pos_tar->x; 
            dy = pos_src->y - pos_tar->y;

	    /* Is source object under the top surface of the target
	     * object?
	     */
	    if(pos_src->z < (pos_tar->z + cb_tar->contact_z_max))
	    {
		/* Source object is under, so do not account for surface. */
	    }
	    else
	    {
		/* Rotate dx and dy about the center of the target
                 * object inversly incase target object has a rotated
                 * heading.
                 */
                dx2 = (cb_tar->cos_heading * dx) +
                    (cb_tar->sin_heading * dy);
                dy2 = (cb_tar->cos_heading * dy) -
                    (cb_tar->sin_heading * dx);

                /* Calculate x and y bounds. */
                x_min = cb_tar->contact_x_min;
                x_max = cb_tar->contact_x_max;
                y_min = cb_tar->contact_y_min;
                y_max = cb_tar->contact_y_max;

                /* Check if rotated dx2 and dy2 are in bounds. */
                if((dx2 >= x_min) && (dx2 <= x_max) &&
                   (dy2 >= y_min) && (dy2 <= y_max)
                )
		    cur_z_height = pos_tar->z + cb_tar->contact_z_max;
	    }
	    break;
	}

	return(cur_z_height);
}


/*
 *	Returns the elevation of (not to) the object of type
 *	ground given the position pos (z coordinate will be
 *	ignored and msl elevation offset will be ignored).
 *
 *	The object must be of type ground, the return value will
 *	be either the ground object's set elevation or from
 *	its heighfield data (if any), whichever is higher.
 */
double SARSimHFGetGroundHeight(
	sar_object_struct *obj_ptr, sar_position_struct *pos
)
{
	double d, dx, dy, drx, dry;
	double cur_z_height = 0.0;
	double sin_theta, cos_theta, x_len, y_len, contact_radius;
	sar_direction_struct *dir_src;
	sar_position_struct *pos_tar = pos, *pos_src;
	sar_object_ground_struct *obj_ground_ptr;


	if((obj_ptr == NULL) ||
           (pos_tar == NULL)
	)
	    return(cur_z_height);

	/* Is this object of type ground? */
	if(obj_ptr->type == SAR_OBJ_TYPE_GROUND)
	    obj_ground_ptr = obj_ptr->data;
	else
	    return(cur_z_height);

	if(obj_ground_ptr == NULL)
	    return(cur_z_height);

	/* Note, here target reffers to the given position and source
	 * reffers to the SAR_OBJ_TYPE_GROUND object.
	 */

	/* Get flat contact radius of ground object. */
	contact_radius = SARSimGetFlatContactRadius(obj_ptr);

	/* All SAR_OBJ_TYPE_GROUND objects have cylendrical solid
	 * surfaces, where the radius is specified by the ground
	 * object's contact bounds (which can be a radius of 0.0).
	 *
	 * Get position of ground object and check distance,
	 * see if outside the ground object's contact radius.
	 */
	pos_src = &obj_ptr->pos;
	/* Calculate deltas, target position to source position.
	 * Note dy needs to be `top left' oriented.
	 */
	dx = pos_tar->x - (pos_src->x + obj_ground_ptr->x_trans);
	dy = (pos_src->y + obj_ground_ptr->y_trans) - pos_tar->y;
	d = hypot(dx, dy);
	if(d <= contact_radius)
	{
	    /* Compare flat elevation first, by checking the ground
	     * object's specified elevation from 0. Note that ground
	     * objects completly ignore the scene structure defined
	     * ground_elevation_msl value.
	     */
	    if((obj_ground_ptr->elevation + pos_src->z +
		obj_ground_ptr->z_trans) > cur_z_height
	    )
		cur_z_height = obj_ground_ptr->elevation + pos_src->z +
		    obj_ground_ptr->z_trans;
	}

	/* Got dx and dy (delta distance of given target position to
	 * source position (the ground object).
	 *
	 * Now rotate dx and dy position based on ground object's
	 * heading, remember that dy is top to bottom.
	 */
	dir_src = &obj_ptr->dir;
	cos_theta = obj_ground_ptr->cos_heading;
	sin_theta = obj_ground_ptr->sin_heading;
	drx = (double)((dx * cos_theta) - (dy * sin_theta));
	dry = (double)((dy * cos_theta) + (dx * sin_theta));

	/* Check if ground object has any heightfield data. */
	x_len = obj_ground_ptr->x_len;
	y_len = obj_ground_ptr->y_len;
	if((obj_ground_ptr->z_point_value != NULL) &&
	   (d < MAX(x_len, y_len))
	)
	{
	    double	half_x_len = obj_ground_ptr->x_len / 2,
			half_y_len = obj_ground_ptr->y_len / 2;

	    if((obj_ground_ptr->grid_x_spacing > 0.0) &&
	       (obj_ground_ptr->grid_y_spacing > 0.0)
	    )
	    {
		/* Calculate point index, remember heightfield
		 * is centered at its origin, so need to add
		 * x and y point half index values to index.
		 */
                int xi = (int)
		    ((drx + half_x_len) / obj_ground_ptr->grid_x_spacing);
                int yi = (int)
		    ((dry + half_y_len) / obj_ground_ptr->grid_y_spacing);

		int grid_points_x = obj_ground_ptr->grid_points_x;
		int grid_points_y = obj_ground_ptr->grid_points_y;
		double grid_x_spacing = obj_ground_ptr->grid_x_spacing;
		double grid_y_spacing = obj_ground_ptr->grid_y_spacing;

		/* Clip xi and yi, oriented `upper left'. */
		if(((xi >= 0) && (xi < grid_points_x)) &&
		   ((yi >= 0) && (yi < grid_points_y))
		)
		{
		    /* The GL version of the heightfield has two
		     * triangles per grid, one on the `upper right' and
		     * one on the `lower left' (see V3DHFLoad*()).
		     *
		     *                  v0 - v1
		     *                  | \   |
		     *                  |  \  |
		     *                  |   \ |
		     *                  v2 - v3
		     */

		    /* Calculate the target position `in the grid' as a
		     * coefficient.
		     */
		    double in_grid_x_c =
			((drx + half_x_len) - (xi * grid_x_spacing)) /
			(((xi + 1) * grid_x_spacing) - (xi * grid_x_spacing));
		    double in_grid_y_c =
			((dry + half_y_len) - (yi * grid_y_spacing)) /
			(((yi + 1) * grid_y_spacing) - (yi * grid_y_spacing));
		    float z_value[4];	/* ul, ur, ll, lr. */
		    float z_result;	/* Calculated height point in grid. */

		    /* At this point we only need to fetch three values
		     * from the ground object since we know we're in one
		     * or the other triangle.
		     */
		    if(in_grid_x_c <= in_grid_y_c)
		    {
			/* In lower left triangle:
			 *
			 * Fetch z values for the lower left three vertices.
			 */

			/* Upper left z value. */
			z_value[0] = (float)(obj_ground_ptr->z_point_value[
			    (yi * grid_points_x) + xi
			]);

			/* Lower left z value. */
			if((yi + 1) < grid_points_y)
			{
			    z_value[2] = (float)(obj_ground_ptr->z_point_value[
				((yi + 1) * grid_points_x) + xi
			    ]);
			}
			else
			{
			    /* No data for lower left, fall back to using upper left. */
			    z_value[2] = z_value[0];
			}

			/* Lower right z value. */
			if(((xi + 1) < grid_points_x) &&
			   ((yi + 1) < grid_points_y)
			)
			{
			    z_value[3] = (float)(obj_ground_ptr->z_point_value[
				((yi + 1) * grid_points_x) + (xi + 1)
			    ]);
			}
			/* No data for lower right, try lower left. */
			else if((yi + 1) < grid_points_y)
			{
			    /* Need to calculate upper right, we don't get it for free. */
			    z_value[3] = z_value[2];
			}
			/* Try upper right. */
			else if((xi + 1) < grid_points_x)
			{
			    /* Need to calculate upper right, we don't get it for free. */
			    z_value[3] = (float)(obj_ground_ptr->z_point_value[
				(yi * grid_points_x) + (xi + 1)
			    ]);
			}
			/* No UR or LL, use z_value[0]. */
			else
			{
			    z_value[3] = z_value[0];
			}

			/* Perform the height interpolation using the
			 * refactored equations. Basically what this does
			 * is only perform the 'differencing' portion
			 * of the interpolation on x and y axis'.
			 * (NOTE: C++ allows floats to be defined mid
			 * code as follows, it's more efficient for
			 * the end code sometimes to do this so it's a
			 * good thing to get used to.)
			 */
                        {
			 float height_delta_x = z_value[3] - z_value[2];
			 float height_delta_y = z_value[0] - z_value[2];
			 z_result = z_value[2] + (float)(height_delta_x *
			    in_grid_x_c) + (float)(height_delta_y * (1.0f-in_grid_y_c));
			}
		    }
		    else
		    {
			/* In upper right triangle:
			 *
			 * Fetch z values for the upper right three vertices.
			 */

			/* Upper left z value. */
			z_value[0] = (float)(obj_ground_ptr->z_point_value[
			    (yi * grid_points_x) + xi
			]);

			/* Upper right z value. */
			if((xi + 1) < grid_points_x)
			{
			    z_value[1] = (float)(obj_ground_ptr->z_point_value[
				(yi * grid_points_x) + (xi + 1)
			    ]);
			}
			else
			{
			    /* No data for upper right, fall back to using upper left. */
			    z_value[1] = z_value[0];
			}

			/* Lower right z value. */
			if(((xi + 1) < grid_points_x) &&
			   ((yi + 1) < grid_points_y)
			)
			{
			    z_value[3] = (float)(obj_ground_ptr->z_point_value[
				((yi + 1) * grid_points_x) + (xi + 1)
			    ]);
			}
			/* No lower right data, try lower left? */
			else if((yi + 1) < grid_points_y)
			{
			    /* Need to calculate lower left, we don't get it for free. */
			    z_value[3] = (float)(obj_ground_ptr->z_point_value[
				((yi + 1) * grid_points_x) + xi
			    ]);
			}
			/* Try upper right? */
			else if((xi + 1) < grid_points_x)
			{
			    z_value[3] = z_value[1];
			}
			/* No UR or LL, use z_value[0]. */
			else
			{
			    z_value[3] = z_value[0];
			}

			/* Perform the height interpolation using the
			 * refactored equations. (Same idea as above, just
			 * different points involved.)
			 */
			{
			 float height_delta_x = z_value[0] - z_value[1];
			 float height_delta_y = z_value[3] - z_value[1];
			 z_result = z_value[1] +
			    (float)(height_delta_x * (1.0-in_grid_x_c)) +
			    (float)(height_delta_y * in_grid_y_c);
			}
		    }

		    /* Is the new interpolated height higher than
		     * cur_z_height?
		     */
		    if(cur_z_height < (z_result + pos_src->z +
			obj_ground_ptr->z_trans
		    ))
			cur_z_height = z_result + pos_src->z +
			    obj_ground_ptr->z_trans;
/*
printf("\r %i(%i) %i(%i) (%.2f %.2f) (%.2f %.2f) %.2f",
 xi, obj_ground_ptr->grid_points_x,
 yi, obj_ground_ptr->grid_points_y,
 in_grid_x_c, in_grid_y_c,
 pos_tar->x, pos_tar->y,
 cur_z_height
); fflush(stdout);
 */
		}

	    }
	}

	return(cur_z_height);
}
