#include <stdio.h> 
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "sfm.h"

#include "sarreality.h"
#include "obj.h"
#include "sarsound.h"
#include "smoke.h"
#include "explosion.h"
#include "messages.h" 
#include "simmanage.h"
#include "simutils.h"
#include "sar.h"
#include "config.h"


double SARSimGetFlatContactRadius(sar_object_struct *obj_ptr);
sar_object_struct *SARSimMatchObjectFromFDM(
        sar_object_struct **list, int total,
        SFMModelStruct *fdm, int *index
);

void SARSimWarpObject(
        sar_scene_struct *scene, sar_object_struct *obj_ptr,
        sar_position_struct *new_pos,
        sar_direction_struct *new_dir
);
int SARSimGetLandingGearState(sar_object_struct *obj_ptr);
void SARSimUpdateLandingGear(sar_obj_landing_gear_struct *lgear_ptr);
void SARSimOpLandingGear(  
        sar_scene_struct *scene, 
        sar_object_struct ***ptr,
        int *total,
        sar_object_struct *obj_ptr,
        int gear_state
);

void SARSimUpdateDoor(
	sar_obj_door_struct *door_ptr
);
void SARSimOpDoor(
        sar_scene_struct *scene,
        sar_object_struct ***ptr,
        int *total,
        sar_object_struct *obj_ptr,
        int state
);

int SARSimGetEngineState(sar_object_struct *obj_ptr);
void SARSimOpEngine(
        sar_scene_struct *scene,
        sar_object_struct ***ptr,
        int *total,
        sar_object_struct *obj_ptr,
        int engine_state               
);

int SARSimGetLightsState(sar_object_struct *obj_ptr);
void SARSimOpLights(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	int state
);
int SARSimGetStrobesState(sar_object_struct *obj_ptr);
void SARSimOpStrobes(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	int state
);
int SARSimGetAttenuate(sar_object_struct *obj_ptr);
void SFMSimOpAttenuate(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	int state
);
  
void SARSimSmokeSpawnCheck(
	sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr,
	time_t life_span
);
int SARSimDoTouchDownCheck(  
	sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr
);
int SARSimDoMidFlightCrash(
        sar_core_struct *core_ptr,
        sar_object_struct *obj_ptr,
        sar_object_struct *obstruction_obj_ptr
);
int SARSimDoPickUpHuman(
	sar_scene_struct *scene,
        sar_object_struct *obj_ptr,
	sar_object_human_struct *obj_human_ptr,
	int human_obj_num
);
int SARSimDoHoistIn(
	sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr
);
void SARSimSetFlyByPosition(
	sar_scene_struct *scene,
        sar_object_struct ***ptr,
        int *total,
        sar_object_struct *obj_ptr,
	sar_position_struct *pos_result
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))


/*
 *      Returns the flat (XY plane) contact bounds radius for
 *      the given object.
 *
 *      Can return 0.0 if contact bounds are not availbale for
 *      the object.
 */
double SARSimGetFlatContactRadius(sar_object_struct *obj_ptr)
{
        double contact_radius = 0.0;
        sar_contact_bounds_struct *cb;

        if(obj_ptr == NULL)
            return(0.0);

        cb = obj_ptr->contact_bounds;
        if(cb == NULL)
            return(0.0);

        switch(cb->contact_shape)
        {
          case SAR_CONTACT_SHAPE_SPHERICAL:
          case SAR_CONTACT_SHAPE_CYLENDRICAL:
            contact_radius = MAX(cb->contact_radius, 0.0);
            break;

          case SAR_CONTACT_SHAPE_RECTANGULAR:
            contact_radius = MAX(
                cb->contact_x_max - cb->contact_x_min,
                cb->contact_y_max - cb->contact_y_min
            );
            break;
        }

        return(MAX(contact_radius, 0.0));
}

/*
 *	Returns the pointer to the object in the give nobject list that
 *	contains an FDM that matches the given fdm or NULL on error.
 *
 *      If index is not NULL then the index to the object number on
 *      the objects list will be set or -1 on failed match.
 */
sar_object_struct *SARSimMatchObjectFromFDM(
        sar_object_struct **list, int total,
        SFMModelStruct *fdm, int *index
)
{
        int i;
        sar_object_struct *obj_ptr;
        sar_object_aircraft_struct *obj_aircraft_ptr;


        if((list == NULL) ||
           (fdm == NULL)
        )
        {
            if(index != NULL)
                (*index) = -1;
                
            return(NULL);
        }
  
        for(i = 0; i < total; i++)
        {
            obj_ptr = list[i]; 
            if(obj_ptr == NULL)
                continue;
        
            switch(obj_ptr->type)
            {
              case SAR_OBJ_TYPE_AIRCRAFT:
                obj_aircraft_ptr = (sar_object_aircraft_struct *)
                    obj_ptr->data;
                if(obj_aircraft_ptr == NULL)
                    break;

                if(obj_aircraft_ptr->fdm == fdm)
                {
                    if(index != NULL)
                        (*index) = i;
                    return(obj_ptr);
                }
                break;

/* Add checking for other object types which may have SFM structures
 * here when needed.
 */

            }
        }

        if(index != NULL)
            (*index) = -1;

        return(NULL);
}


/*
 *	Updates all position and attitude values for the object to
 *	the newly specified position and values.
 *
 *	If any input is NULL then the values will be left unchanged.
 *
 *	If the given input pointers match the pointers of the structures
 *	on the object then the structure in question will be left
 *	unchanged.
 *
 *	Rotatation values for the contact bounds and ground object
 *	structure will be updated as needed.
 */
void SARSimWarpObject(
        sar_scene_struct *scene, sar_object_struct *obj_ptr,  
        sar_position_struct *new_pos,
	sar_direction_struct *new_dir
)
{
	sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	sar_object_ground_struct *obj_ground_ptr = NULL;
	sar_contact_bounds_struct *cb;
	SFMModelStruct *fdm = NULL;


	if((scene == NULL) ||
           (obj_ptr == NULL)
	)
	    return;

	if((new_pos != NULL) &&
           (new_pos != &obj_ptr->pos)
	)
	{
	    obj_ptr->pos.x = new_pos->x;
            obj_ptr->pos.y = new_pos->y;
            obj_ptr->pos.z = new_pos->z;
	}
	if((new_dir != NULL) &&
           (new_dir != &obj_ptr->dir)
	)
	{
	    obj_ptr->dir.heading = new_dir->heading;
            obj_ptr->dir.pitch = new_dir->pitch;
            obj_ptr->dir.bank = new_dir->bank;
	}

	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
	    if(obj_aircraft_ptr != NULL)
	    {
		/* Get pointer to SFM flight dynamics structure. */
		fdm = obj_aircraft_ptr->fdm;
	    }
	    break;

	  case SAR_OBJ_TYPE_GROUND:
	    obj_ground_ptr = (sar_object_ground_struct *)obj_ptr->data;
	    if(obj_ground_ptr != NULL)
	    {
		/* Update trig functions for ground's heightfield
		 * rotation.
		 */
		if(new_dir != NULL)
		{
		    obj_ground_ptr->cos_heading = cos(-new_dir->heading);
                    obj_ground_ptr->sin_heading = sin(-new_dir->heading);
		}
	    }
	    break;
	}


	/* Set values on FDM if exists. */
	if(fdm != NULL)
	{
	    if(new_pos != NULL)
	    {
		fdm->position.x = new_pos->x;
                fdm->position.y = new_pos->y;
                fdm->position.z = new_pos->z;
	    }
            if(new_dir != NULL)
            {
                fdm->direction.heading = new_dir->heading;
                fdm->direction.pitch = new_dir->pitch;
                fdm->direction.bank = new_dir->bank;
            }
	}

	/* Update values on contact bounds. */
	cb = obj_ptr->contact_bounds;
	if(cb != NULL)
	{
	    /* Update values based on contact shape. */
	    if(cb->contact_shape == SAR_CONTACT_SHAPE_RECTANGULAR)
	    {
		/* For rectangular contact bounds, the cos and sin
		 * values for the heading need to be updated based
		 * on the new rotation.
		 */
		if(new_dir != NULL)
		{
		    cb->cos_heading = cos(-new_dir->heading);
                    cb->sin_heading = sin(-new_dir->heading);
		}
	    }

	}

	return;
}

/*
 *	Gets the landing gear state of the first landing gear
 *	on the object. Returns true if gear is down, false if up.
 */
int SARSimGetLandingGearState(sar_object_struct *obj_ptr)
{
        sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_obj_landing_gear_struct *lgear_ptr = NULL;


        if(obj_ptr == NULL)
            return(0);

        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = obj_ptr->data;

	    if(obj_aircraft_ptr->total_landing_gears > 0)
		lgear_ptr = obj_aircraft_ptr->landing_gear[0];
            break;
        }

	if(lgear_ptr == NULL)
	    return(0);
	else
	    return(lgear_ptr->flags & SAR_LGEAR_FLAG_STATE);
}

/*
 *	Updates animated position of landing gear.
 */
void SARSimUpdateLandingGear(sar_obj_landing_gear_struct *lgear_ptr)
{
	sar_obj_flags_t flags;


	if(lgear_ptr == NULL)
	    return;

	flags = lgear_ptr->flags;

	if(flags & SAR_LGEAR_FLAG_STATE)
	{
	    /* Raising. */

            if(lgear_ptr->anim_pos == 0)
                return;

	    if((flags & SAR_LGEAR_FLAG_DAMAGED) ||
               (flags & SAR_LGEAR_FLAG_MISSING) ||
               (flags & SAR_LGEAR_FLAG_FIXED)
	    )
		return;

            lgear_ptr->anim_pos = (sar_grad_anim_t)MAX(
                (int)lgear_ptr->anim_pos -
                ((int)lgear_ptr->anim_rate * time_compensation *
                    time_compression),
                0
            );
	}
	else
	{
	    /* Lowering. */

            if(lgear_ptr->anim_pos == (sar_grad_anim_t)-1)
                return;

            if((flags & SAR_LGEAR_FLAG_DAMAGED) ||
               (flags & SAR_LGEAR_FLAG_MISSING)
            )
                return;

            lgear_ptr->anim_pos = (sar_grad_anim_t)MIN(
                (int)lgear_ptr->anim_pos +
                ((int)lgear_ptr->anim_rate * time_compensation *
                    time_compression),
                (sar_grad_anim_t)-1   
            );   
	}

	return;
}

/*
 *	Raise or lower landing gear on object.
 */
void SARSimOpLandingGear(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	int gear_state
)
{
	int i;
        sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_obj_landing_gear_struct *lgear_ptr;


        if(obj_ptr == NULL)
            return;

        switch(obj_ptr->type)
        {
	  case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

	    /* No landing gears on object? */
	    if(obj_aircraft_ptr->total_landing_gears <= 0)
		break;

	    /* Cannot operate landing gears while landed. */
	    if(obj_aircraft_ptr->landed)
		break;

	    /* Get pointer to first landing gear. */
	    lgear_ptr = obj_aircraft_ptr->landing_gear[0];
	    if(lgear_ptr == NULL)
		break;

	    /* Gear is up, we want to lower it? */
	    if((lgear_ptr->flags & SAR_LGEAR_FLAG_STATE) &&
	       !gear_state
	    )
	    {
		/* Landing gear is currently up, we want to lower it. */

                for(i = 0; i < obj_aircraft_ptr->total_landing_gears; i++)
                {
                    lgear_ptr = obj_aircraft_ptr->landing_gear[i];
                    if(lgear_ptr == NULL)
                        continue;

                    if((lgear_ptr->flags & SAR_LGEAR_FLAG_DAMAGED) ||
		       (lgear_ptr->flags & SAR_LGEAR_FLAG_MISSING)
		    )
                        continue;

		    lgear_ptr->flags &= ~SAR_LGEAR_FLAG_STATE;
		}
	    }
            /* Gear is down, we want to raise it? */
	    else if(!(lgear_ptr->flags & SAR_LGEAR_FLAG_STATE) &&
                    gear_state
            )
            {
                /* Landing gear is currently down, we want to raise it. */

		for(i = 0; i < obj_aircraft_ptr->total_landing_gears; i++)
		{
		    lgear_ptr = obj_aircraft_ptr->landing_gear[i];
		    if(lgear_ptr == NULL)
			continue;

		    if((lgear_ptr->flags & SAR_LGEAR_FLAG_FIXED) ||
                       (lgear_ptr->flags & SAR_LGEAR_FLAG_MISSING) ||
                       (lgear_ptr->flags & SAR_LGEAR_FLAG_DAMAGED)
                    )
			continue;

                    lgear_ptr->flags |= SAR_LGEAR_FLAG_STATE;
		}
	    }
	    break;
	}

	return;	
}

/*
 *      Updates animated position of door.
 */
void SARSimUpdateDoor(
	sar_obj_door_struct *door_ptr
)
{
        sar_obj_flags_t flags;


        if(door_ptr == NULL)
            return;
            
        flags = door_ptr->flags;

        if(flags & SAR_DOOR_FLAG_STATE)
        {
            /* Opening. */

            if(door_ptr->anim_pos == (sar_grad_anim_t)-1)
                return;

            if(flags & SAR_DOOR_FLAG_FIXED)
                return;

            door_ptr->anim_pos = (sar_grad_anim_t)MIN(
		(int)door_ptr->anim_pos +
                ((int)door_ptr->anim_rate * time_compensation *
                    time_compression),
                (sar_grad_anim_t)-1
            );
        }
        else
        {
            /* Closing. */

            if(door_ptr->anim_pos == 0)
                return;

            if(flags & SAR_DOOR_FLAG_FIXED)
                return;

            door_ptr->anim_pos = (sar_grad_anim_t)MAX(
		(int)door_ptr->anim_pos -
                ((int)door_ptr->anim_rate * time_compensation *
                    time_compression),
                0
            );
        }

	return;
}

/*
 *	Open or close door on object.
 */
void SARSimOpDoor(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	int state
)
{
        sar_object_aircraft_struct *obj_aircraft_ptr;
        sar_obj_door_struct *door_ptr;


        if(obj_ptr == NULL)
            return;

        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)
                obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

	    door_ptr = &obj_aircraft_ptr->rescue_door;


            /* Door is opened, we want to close it? */
            if((door_ptr->flags & SAR_DOOR_FLAG_STATE) &&
               !state
            )
            {
                if(!(door_ptr->flags & SAR_DOOR_FLAG_FIXED))
                    door_ptr->flags &= ~SAR_DOOR_FLAG_STATE;
            }
            /* Door is closed, we want to open it? */
            else if(!(door_ptr->flags & SAR_DOOR_FLAG_STATE) &&
                    state
            )
            {
                if(!(door_ptr->flags & SAR_DOOR_FLAG_FIXED) &&
                   !(door_ptr->flags & SAR_DOOR_FLAG_LOCKED)
		)
		    door_ptr->flags |= SAR_DOOR_FLAG_STATE;
            }
            break;
        }

	return;
}

/*
 *	Returns the engine state of the object.
 *
 *	If object is not a type that has engines, then
 *	SAR_ENGINE_STATE_OFF is returned.
 */
int SARSimGetEngineState(sar_object_struct *obj_ptr)
{
	sar_object_aircraft_struct *obj_aircraft_ptr;


	if(obj_ptr == NULL)
	    return(SAR_ENGINE_STATE_OFF);

	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = obj_ptr->data;
	    return(obj_aircraft_ptr->engine_state);
	    break;
	}

	return(SAR_ENGINE_STATE_OFF);
}

/*
 *	Turns engine off or initialize (on).
 */
void SARSimOpEngine(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	int engine_state
)
{
        sar_object_aircraft_struct *obj_aircraft_ptr;
  
 
        if(obj_ptr == NULL)
            return;

        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)
                obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

	    if(obj_aircraft_ptr->air_worthy_state != SAR_AIR_WORTHY_FLYABLE)
		break;

	    if(engine_state == SAR_ENGINE_STATE_ON)
	    {
		obj_aircraft_ptr->engine_state = SAR_ENGINE_STATE_INIT;
		obj_aircraft_ptr->next_engine_on = cur_millitime +
		    2000;
	    }
	    else
	    {
		obj_aircraft_ptr->engine_state = SAR_ENGINE_STATE_OFF;
		obj_aircraft_ptr->next_engine_on = 0;
	    }
	    break;
	}

	return;
}

/*
 *	Returns the state of the first light structure on the object.
 *
 *	Can return false if no lights exist or error.
 */
int SARSimGetLightsState(sar_object_struct *obj_ptr)
{
        int i;
        sar_light_struct *light_ptr;


        if(obj_ptr == NULL)
            return(0);

        for(i = 0; i < obj_ptr->total_lights; i++)
        {
            light_ptr = obj_ptr->light[i];
            if(light_ptr == NULL)
                continue;

            /* Is it a strobe light? */
            if(!(light_ptr->flags & SAR_LIGHT_FLAG_STROBE) &&
               !(light_ptr->flags & SAR_LIGHT_FLAG_ATTENUATE)
	    )
                return(light_ptr->flags & SAR_LIGHT_FLAG_ON);
        }

        return(0);
}

/*
 *	Turns all regular lights on object on or off.
 */
void SARSimOpLights(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
        int *total,
        sar_object_struct *obj_ptr,
        int state
)
{
	int i;
	sar_light_struct *light_ptr;

	if(obj_ptr == NULL)
            return;

	for(i = 0; i < obj_ptr->total_lights; i++)
	{
	    light_ptr = obj_ptr->light[i];
	    if(light_ptr == NULL)
		continue;

	    /* Skip non-standard lights. */
	    if((light_ptr->flags & SAR_LIGHT_FLAG_STROBE) ||
               (light_ptr->flags & SAR_LIGHT_FLAG_ATTENUATE)
	    )
		continue;

	    if(state)
	    {
		/* Turn on. */
		light_ptr->flags |= SAR_LIGHT_FLAG_ON;
	    }
	    else
	    {
		/* Turn off. */
		light_ptr->flags &= ~SAR_LIGHT_FLAG_ON;
	    }
	}

	return;
}

/*
 *      Returns the state of the first strobe light structure on the
 *	object.
 *
 *      Can return false if no lights exist or error.
 */
int SARSimGetStrobesState(sar_object_struct *obj_ptr)
{
	int i;
	sar_light_struct *light_ptr;


	if(obj_ptr == NULL)
	    return(0);

	for(i = 0; i < obj_ptr->total_lights; i++)
	{
	    light_ptr = obj_ptr->light[i];
	    if(light_ptr == NULL)
		continue;

            /* Is it a strobe light? */
            if(light_ptr->flags & SAR_LIGHT_FLAG_STROBE)
                return(light_ptr->flags & SAR_LIGHT_FLAG_ON);
	}

	return(0);
}

/*
 *	Turns all strobe lights on object on or off.
 */
void SARSimOpStrobes(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	int state
)
{
        int i;
        sar_light_struct *light_ptr;

        if(obj_ptr == NULL)
            return;

        for(i = 0; i < obj_ptr->total_lights; i++)
        {
            light_ptr = obj_ptr->light[i];
            if(light_ptr == NULL)
                continue;

            /* Skip non-strobe lights. */
            if(!(light_ptr->flags & SAR_LIGHT_FLAG_STROBE))
                continue;

            if(state)
            {
                /* Turn on. */
                light_ptr->flags |= SAR_LIGHT_FLAG_ON;
            }
            else
            {   
                /* Turn off. */
                light_ptr->flags &= ~SAR_LIGHT_FLAG_ON;
            }
        }

        return;
}

/*
 *      Returns the state of the first attenuation light structure on the
 *      object.
 *
 *      Can return false if no lights exist or error.
 */
int SARSimGetAttenuate(sar_object_struct *obj_ptr)
{
        int i;
        sar_light_struct *light_ptr;


        if(obj_ptr == NULL)
            return(0);

        for(i = 0; i < obj_ptr->total_lights; i++)
        {
            light_ptr = obj_ptr->light[i];
            if(light_ptr == NULL)  
                continue;

            /* Is it a attenuation light? */
            if(light_ptr->flags & SAR_LIGHT_FLAG_ATTENUATE)
                return(light_ptr->flags & SAR_LIGHT_FLAG_ON);
        }

        return(0);
}

/*
 *      Turns all attenuation lights on object on or off.
 */
void SFMSimOpAttenuate(
        sar_scene_struct *scene,
        sar_object_struct ***ptr,
        int *total,
        sar_object_struct *obj_ptr,
        int state
)
{
        int i;
        sar_light_struct *light_ptr;

        if(obj_ptr == NULL)
            return;

        for(i = 0; i < obj_ptr->total_lights; i++)
        {
            light_ptr = obj_ptr->light[i];
            if(light_ptr == NULL)
                continue;

            /* Skip non-attenuation lights. */
            if(!(light_ptr->flags & SAR_LIGHT_FLAG_ATTENUATE))
                continue;

            if(state)
            {
                /* Turn on. */
                light_ptr->flags |= SAR_LIGHT_FLAG_ON;
            }
            else
            {
                /* Turn off. */
                light_ptr->flags &= ~SAR_LIGHT_FLAG_ON;
            }
        }

        return;
}


/*
 *	Checks if object needs to generate a smoke puff.
 *
 *	Calls to this function must be timmed properly!
 */
void SARSimSmokeSpawnCheck(
	sar_core_struct *core_ptr,
        sar_object_struct *obj_ptr,
	time_t life_span
)
{
	int generate = 0;
	double contact_radius;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_object_explosion_struct *obj_explosion_ptr;


        if((core_ptr == NULL) ||
           (obj_ptr == NULL)
	)
            return;

	contact_radius = SARSimGetFlatContactRadius(obj_ptr);

        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)
                obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

	    /* Generate smoke if out of control. */
	    if(obj_aircraft_ptr->air_worthy_state ==
		SAR_AIR_WORTHY_OUT_OF_CONTROL
	    )
		generate = 1;

	    /* Generate light smoke puff? */
	    if(generate)
		SmokeCreatePosition(
		    core_ptr->scene,
		    &core_ptr->object, &core_ptr->total_objects,
		    &obj_ptr->pos,
		    contact_radius * 0.5,
		    SAR_STD_TEXNAME_SMOKE_LIGHT,
                    life_span
		);
	    break;

	  case SAR_OBJ_TYPE_FIRE:

	    break;

          case SAR_OBJ_TYPE_EXPLOSION:
	    obj_explosion_ptr = (sar_object_explosion_struct *)obj_ptr->data;
	    if(obj_explosion_ptr == NULL)
		break;

	    generate = 1;

            if(generate)
                SmokeCreatePosition(
                    core_ptr->scene,
                    &core_ptr->object, &core_ptr->total_objects,
                    &obj_ptr->pos,
                    obj_explosion_ptr->radius * 0.8,
                    SAR_STD_TEXNAME_SMOKE_DARK,
                    life_span
                );
            break;
	}

	return;
}

/*
 *      Moves the human object specified by obj_human_ptr (substructure
 *	only) into the object specified by obj_ptr's rescue basket.
 *
 *	Note that obj_human_ptr may not be the substructure of obj_ptr.
 *
 *	All inputs assumed valid.
 *
 *      Returns 1 if the human object was actually picked up, or 0 on
 *	failure.
 */
int SARSimDoPickUpHuman(
	sar_scene_struct *scene,
        sar_object_struct *obj_ptr,
        sar_object_human_struct *obj_human_ptr,
	int human_obj_num
)
{
	int n;
	int passengers = 0, passengers_max = 0;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_obj_hoist_struct *hoist_ptr = NULL;


	if(scene == NULL)
	    return(0);

	/* Handle by object type. */
	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
	    if(obj_aircraft_ptr != NULL)
	    {
		hoist_ptr = &obj_aircraft_ptr->hoist;
		passengers = obj_aircraft_ptr->passengers;
		passengers_max = obj_aircraft_ptr->passengers_max;
	    }
	    break;
	}

	/* Did we get a hoist structure? */
	if(hoist_ptr == NULL)
	    return(0);

	/* No more room left onboard for additional passengers? */
	if(passengers >= passengers_max)
	{
	    /* Warn if this is the player object and the scene is
	     * not currently displaying any messages.
	     */
	    if((obj_ptr == scene->player_obj_ptr) &&
               (scene->message_display_untill <= 0)
	    )
	    {
		SARMessageAdd(
		    scene,
		    SAR_MESG_NO_ROOM_LEFT_FOR_PASSENGERS
		);
	    }
	    return(0);
	}

	/* Can hoist accomidate more (just limit to 1 for now)? */
	if(hoist_ptr->total_occupants >= 1)
	    return(0);


	/* Add human to hoist structure's list of occupants. */
	if(hoist_ptr->total_occupants < 0)
	    hoist_ptr->total_occupants = 0;

	n = hoist_ptr->total_occupants;
	hoist_ptr->total_occupants++;
	hoist_ptr->occupant = (int *)realloc(
	    hoist_ptr->occupant,
	    hoist_ptr->total_occupants * sizeof(int)
	);
	if(hoist_ptr->occupant == NULL)
	{
	    hoist_ptr->total_occupants = 0;
	    return(0);
	}
	else
	{
	    hoist_ptr->occupant[n] = human_obj_num;
	}

	/* Update nessesary flags on human object. */
	obj_human_ptr->flags = ~SAR_HUMAN_FLAG_NEED_RESCUE;
        obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_SIT;
        obj_human_ptr->flags |= SAR_HUMAN_FLAG_SIT_DOWN;
        obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_SIT_UP;
        obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_LYING;
	/* Leave SAR_HUMAN_FLAG_ALERT and SAR_HUMAN_FLAG_AWARE as is. */
        obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_IN_WATER;
	obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_RUN_TOWARDS;
	obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_RUN_AWAY;
	obj_human_ptr->flags |= SAR_HUMAN_FLAG_GRIPPED;

 	return(1);
}


/*
 *	This function should be called each time the hoist is pulled
 *	into the aircraft. It will delete any objects referanced as
 *	occupants in the aircraft's hoist and update the
 *	passenger count.
 *
 *	Inputs are assumed valid, returns 1 if any objects were deleted.
 */
int SARSimDoHoistIn(
	sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr
)
{
	sar_object_aircraft_struct *obj_aircraft_ptr;
	int i, n;

	int is_player = 0;
	int prev_passengers = 0;
	int *cur_passengers = NULL, *passengers_max = NULL;
	sar_obj_hoist_struct *hoist_ptr = NULL;
	sar_object_struct *passenger_obj_ptr;
	sar_object_human_struct *obj_human_ptr;
	sar_scene_struct *scene;


	scene = core_ptr->scene;
	if(scene == NULL)
	    return(0);

	/* Is this the player object? */
	if(scene->player_obj_ptr == obj_ptr)
	    is_player = 1;

	/* Handle by object type. */
	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
	    obj_aircraft_ptr = obj_ptr->data;
	    if(obj_aircraft_ptr != NULL)
	    {
		prev_passengers = obj_aircraft_ptr->passengers;
		cur_passengers = &obj_aircraft_ptr->passengers;
		passengers_max = &obj_aircraft_ptr->passengers_max;
		hoist_ptr = &obj_aircraft_ptr->hoist;
	    }
	    break;
	}

	/* If object has no hoist then skip. */
	if(hoist_ptr == NULL)
	    return(0);

	/* Delete all objects in hoist. */
	for(i = 0; i < hoist_ptr->total_occupants; i++)
	{
	    n = hoist_ptr->occupant[i];
	    if(SARObjIsAllocated(
                core_ptr->object,
                core_ptr->total_objects,
                n
            ))
                passenger_obj_ptr = core_ptr->object[n];
            else
                continue;

	    /* Do not delete the object that owns the hoist
	     * structure. Skip it entirly.
	     */
	    if(passenger_obj_ptr == obj_ptr)
		continue;

	    /* Handle by object type. */
	    switch(passenger_obj_ptr->type)
	    {
	      case SAR_OBJ_TYPE_HUMAN:
		obj_human_ptr = passenger_obj_ptr->data;
		if(obj_human_ptr != NULL)
		{
		    /* If this is the player object, then print `hoist-in
		     * message' specified on the human object.
		     */
		    if(is_player)
			SARMessageAdd(scene, obj_human_ptr->mesg_hoist_in);
		}
		break;
	    }

	    /* Delete object in the hoist's rescue basket. */
	    SARObjDelete(
		core_ptr,
		&core_ptr->object,
		&core_ptr->total_objects,
		n
	    );

	    /* Increment passenger count on aircraft. */
	    if(cur_passengers != NULL)
		(*cur_passengers) = (int)((*cur_passengers) + 1);
	}

	/* Delete occupants list. */
	free(hoist_ptr->occupant);
	hoist_ptr->occupant = NULL;
	hoist_ptr->total_occupants = 0;


	/* Print new number of passengers on aircraft. */
	if((cur_passengers != NULL) &&
	   (passengers_max != NULL)
	)
	{
	    /* Got more passengers? */
	    if(prev_passengers < (*cur_passengers))
	    {
		/* Is this the player object? */
	        if(is_player)
	        {
		    char text[256];

		    sprintf(
			text,
			"Passengers: %i(%i)",
			(int)(*cur_passengers),
			(int)(*passengers_max)
		    );
		    SARMessageAdd(scene, text);
		}
		else
		{
		    char name[SAR_OBJ_NAME_MAX];
		    char text[SAR_OBJ_NAME_MAX + 256];

		    if(obj_ptr->name == NULL)
		    {
			int obj_num = SARGetObjectNumberFromPointer(
			    scene, core_ptr->object, core_ptr->total_objects,
			    obj_ptr
			);
			if(SAR_OBJ_NAME_MAX > 80)
			    sprintf(name, "Object #%i\n", obj_num);
			else
			    (*name) = '\0';
		    }
		    else
		    {
			strncpy(name, obj_ptr->name, SAR_OBJ_NAME_MAX);
		    }
		    name[SAR_OBJ_NAME_MAX - 1] = '\0';

		    sprintf(text, "%s picked up %i passengers",
			name, (int)((*cur_passengers) - prev_passengers)
		    );
		    SARMessageAdd(scene, text);
		}
	    }
	}

	return(0);
}


/*
 *	Calculates the `fly by view' position with respect to
 *	the given object. The result of the calculation is stored
 *	in the pos_result structure.
 */
void SARSimSetFlyByPosition(
	sar_scene_struct *scene,
	sar_object_struct ***ptr,
	int *total,
	sar_object_struct *obj_ptr,
	sar_position_struct *pos_result
)
{
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_direction_struct vel_dir;
	sar_position_struct *vel, *pos;
	double r, h, d, contact_radius;


	if((obj_ptr == NULL) ||
	   (pos_result == NULL)
	)
	    return;

        contact_radius = SARSimGetFlatContactRadius(obj_ptr);

	switch(obj_ptr->type)
	{
	  case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

	    pos = &obj_ptr->pos;
	    vel = &obj_aircraft_ptr->vel;

	    d = (5 * SAR_CYCLE_TO_SEC_COEFF) *
		obj_aircraft_ptr->speed;
	    if(d < (contact_radius * 2))
		d = contact_radius * 2;

            h = (5 * SAR_CYCLE_TO_SEC_COEFF) * vel->z;

	    r = sqrt(MAX((d * d) - (h * h), 0.0));
	    if(r < (contact_radius * 2))
                r = contact_radius * 2; 

	    vel_dir.heading = SFMSanitizeRadians(
	        ((PI / 2) - atan2(vel->y, vel->x)) +
		obj_ptr->dir.heading
	    );
	    pos_result->x = (r * sin(vel_dir.heading)) + pos->x -
		(contact_radius * 2);
	    pos_result->y = (r * cos(vel_dir.heading)) + pos->y -
		(contact_radius * 2);
	    pos_result->z = h + pos->z;

	    /* Don't position lower than object's center to ground. */
	    if(pos_result->z < (pos->z + obj_aircraft_ptr->center_to_ground_height))
		pos_result->z = pos->z + obj_aircraft_ptr->center_to_ground_height;

	    break;

	  default:

	    break;
	}

	return;
}

